"""
Copyright (c) Cutleast
"""

from __future__ import annotations

from dataclasses import dataclass
from io import BufferedReader

from .datatypes import Flags, Integer


@dataclass
class Header:
    """
    Class for archive header.

    See here for more information:
        https://en.uesp.net/wiki/Skyrim_Mod:Archive_File_Format#Header
    """

    class ArchiveFlags(Flags):
        IncludeDirectoryNames = 0x1
        """
        Include Directory Names. (The game may not load a BSA without this bit set.)
        Set in all official BSA files.
        """

        IncludeFileNames = 0x2
        """
        Include File Names. (The game may not load a BSA without this bit set.)
        Set in all official BSA files.
        """

        CompressedArchive = 0x4
        """
        Compressed Archive. This does not mean all files are compressed.
        It means they are compressed by default.
        """

        RetainDirectoryNames = 0x8
        """
        Retain Directory Names. Has no effect on the file structure.
        """

        RetainFileNames = 0x10
        """
        Retain File Names. Has no effect on the file structure.
        """

        RetainFileNameOffsets = 0x20
        """
        Retain File Name Offsets. Has no effect on the file structure.
        """

        Xbox360archive = 0x40
        """
        Xbox360 archive. Hash values and numbers after the header are encoded big-endian.
        """

        RetainStringsDuringStartup = 0x80
        """
        Retain Strings During Startup. Has no effect on the file structure.
        """

        EmbedFileNames = 0x100
        """
        Embed File Names. Indicates the file data blocks begin with a bstring containing
        the full path of the file. For example, in "Skyrim - Textures.bsa" the first data
        block is `$2B textures\\effects\\fxfluidstreamdripatlus.dds`
        ($2B indicating the name is 43 bytes).
        The data block begins immediately after the bstring.
        """

        XMemCodec = 0x200
        """
        XMem Codec. This can only be used with Bit 3 (Compress Archive).
        This is an Xbox 360 only compression algorithm.
        """

    class FileFlags(Flags):
        Meshes = 0x1
        Textures = 0x2
        Menus = 0x4
        Sounds = 0x8
        Voices = 0x10
        Shaders = 0x20
        Trees = 0x40
        Fonts = 0x80
        Miscellaneous = 0x100

    file_id: bytes = b"BSA\x00"
    version: int = 0x69  # Skyrim SE as default version
    offset: int = 0x24  # Header has fix size of 36 bytes
    """
    Offset of beginning of folder records. All headers are the same size,
    therefore this value is 36 (0x24).
    """

    archive_flags: ArchiveFlags = ArchiveFlags(
        ArchiveFlags.IncludeDirectoryNames
        | ArchiveFlags.IncludeFileNames
        | ArchiveFlags.CompressedArchive
        | ArchiveFlags.RetainDirectoryNames
        | ArchiveFlags.RetainFileNames
        | ArchiveFlags.RetainFileNameOffsets
    )
    folder_count: int = 0
    """
    Count of all folders in archive.
    """

    file_count: int = 0
    """
    Count of all files in archive.
    """

    total_folder_name_length: int = 0
    """
    Total length of all folder names, including \0's but not including
    the prefixed length byte.
    """

    total_file_name_length: int = 0
    """
    Total length of all file names, including \0's.
    """

    file_flags: FileFlags = FileFlags(0)
    padding: int = 0

    @staticmethod
    def parse(stream: BufferedReader) -> Header:
        """
        Parses a header from a stream.

        Args:
            stream (BufferedReader): Stream to parse from.

        Raises:
            ValueError: when the archive version is not supported

        Returns:
            Header: Parsed header
        """

        file_id: bytes = stream.read(4)
        version: int = Integer.parse(stream, Integer.IntType.ULong)

        if version != 105:
            raise ValueError("Archive version is not supported!")

        return Header(
            file_id=file_id,
            version=version,
            offset=Integer.parse(stream, Integer.IntType.ULong),
            archive_flags=Header.ArchiveFlags.parse(stream, Integer.IntType.ULong),
            folder_count=Integer.parse(stream, Integer.IntType.ULong),
            file_count=Integer.parse(stream, Integer.IntType.ULong),
            total_folder_name_length=Integer.parse(stream, Integer.IntType.ULong),
            total_file_name_length=Integer.parse(stream, Integer.IntType.ULong),
            file_flags=Header.FileFlags.parse(stream, Integer.IntType.UShort),
            padding=Integer.parse(stream, Integer.IntType.UShort),
        )

    def dump(self) -> bytes:
        """
        Dumps the header to a byte array.

        Returns:
            bytes: Dumped byte array.
        """

        data: bytes = b""

        data += self.file_id
        data += Integer.dump(self.version, Integer.IntType.ULong)
        data += Integer.dump(self.offset, Integer.IntType.ULong)
        data += self.archive_flags.dump(Integer.IntType.ULong)
        data += Integer.dump(self.folder_count, Integer.IntType.ULong)
        data += Integer.dump(self.file_count, Integer.IntType.ULong)
        data += Integer.dump(self.total_folder_name_length, Integer.IntType.ULong)
        data += Integer.dump(self.total_file_name_length, Integer.IntType.ULong)
        data += self.file_flags.dump(Integer.IntType.UShort)
        data += Integer.dump(self.padding, Integer.IntType.UShort)

        return data



"""
Copyright (c) Cutleast
"""

import os
import struct
from enum import Enum, IntFlag, auto
from typing import Literal, Optional, Self, overload

from .utilities import Stream, get_stream, read_data


class Integer:
    """
    Class for all types of signed and unsigned integers.
    """

    class IntType(Enum):
        UInt8 = (1, False)
        """Unsigned Integer of size 1."""

        UInt16 = (2, False)
        """Unsigned Integer of size 2."""

        UInt32 = (4, False)
        """Unsigned Integer of size 4."""

        UInt64 = (8, False)
        """Unsigned Integer of size 8."""

        UShort = (2, False)
        """Same as UInt16."""

        ULong = (4, False)
        """Same as UInt32."""

        Int8 = (1, True)
        """Signed Integer of Size 1."""

        Int16 = (2, True)
        """Signed Integer of Size 2."""

        Int32 = (4, True)
        """Signed Integer of Size 4."""

        Int64 = (8, True)
        """Signed Integer of Size 8."""

        Short = (2, True)
        """Same as Int16."""

        Long = (4, True)
        """Same as Int32."""

    @staticmethod
    def parse(data: Stream | bytes, type: IntType) -> int:
        """
        Parses an integer of the specified type from the specified data.

        Args:
            data (Stream | bytes): Stream or byte array.
            type (IntType): Integer type.

        Returns:
            int: Parsed integer.
        """

        size: int
        signed: bool
        size, signed = type.value

        return int.from_bytes(
            get_stream(data).read(size), byteorder="little", signed=signed
        )

    @staticmethod
    def dump(value: int, type: IntType) -> bytes:
        """
        Dumps an integer of the specified type to a byte array.

        Args:
            value (int): Integer.
            type (IntType): Integer type.

        Returns:
            bytes: Byte array.
        """

        size: int
        signed: bool
        size, signed = type.value

        return value.to_bytes(size, byteorder="little", signed=signed)


class String:
    """
    Class for all types of chars and strings.
    """

    ENCODING: str = "cp1252"
    """The default encoding used in Bethesda's file formats."""

    class StrType(Enum):
        Char = auto()
        """8-bit character."""

        WChar = auto()
        """16-bit character."""

        String = auto()
        """Not-terminated string."""

        WString = auto()
        """Not-terminated string prefixed by UInt16."""

        BZString = auto()
        """Null-terminated string prefixed by UInt8."""

        BString = auto()
        """Not-terminated string prefixed by UInt8."""

        List = auto()
        """List of strings separated by `\\x00`."""

    @staticmethod
    @overload
    def parse(
        data: Stream | bytes, type: Literal[StrType.Char], size: Literal[1] = 1
    ) -> str: ...

    @staticmethod
    @overload
    def parse(
        data: Stream | bytes, type: Literal[StrType.WChar], size: Literal[1] = 1
    ) -> str: ...

    @staticmethod
    @overload
    def parse(
        data: Stream | bytes, type: Literal[StrType.String], size: int
    ) -> str: ...

    @staticmethod
    @overload
    def parse(
        data: Stream | bytes, type: Literal[StrType.WString], size: None = None
    ) -> str: ...

    @staticmethod
    @overload
    def parse(
        data: Stream | bytes, type: Literal[StrType.BString], size: None = None
    ) -> str: ...

    @staticmethod
    @overload
    def parse(
        data: Stream | bytes, type: Literal[StrType.BZString], size: None = None
    ) -> str: ...

    @staticmethod
    @overload
    def parse(
        data: Stream | bytes, type: Literal[StrType.List], size: int
    ) -> list[str]: ...

    @staticmethod
    def parse(
        data: Stream | bytes, type: StrType, size: Optional[int] = None
    ) -> list[str] | str:
        """
        Parses a string of the specified type from the specified data.

        Args:
            data (Stream | bytes): Stream or byte array.
            type (StrType): String type.
            size (Optional[int], optional): Size of the string(s). Defaults to None.

        Returns:
            list[str] | str: Parsed string(s).
        """

        stream = get_stream(data)

        match type:
            case type.Char:
                text = read_data(stream, 1)

            case type.WChar:
                text = read_data(stream, 2)

            case type.String:
                if size is None:
                    raise ValueError("'size' must not be None when 'type' is 'String'!")

                text = read_data(stream, size)

            case type.WString:
                size = Integer.parse(stream, Integer.IntType.UInt16)
                text = read_data(stream, size)

            case type.BZString | type.BString:
                size = Integer.parse(stream, Integer.IntType.UInt8)
                text = read_data(stream, size).strip(b"\x00")

            case type.List:
                strings: list[str] = []

                if size is None:
                    raise ValueError("'size' must not be None when 'type' is 'List'!")

                while len(strings) < size:
                    string = b""
                    while (char := stream.read(1)) != b"\x00" and char:
                        string += char

                    if string:
                        strings.append(string.decode(String.ENCODING))

                return strings

        return text.decode()

    @staticmethod
    @overload
    def dump(value: list[str], type: Literal[StrType.List]) -> bytes: ...

    @staticmethod
    @overload
    def dump(
        value: str,
        type: Literal[
            StrType.Char,
            StrType.WChar,
            StrType.String,
            StrType.WString,
            StrType.BString,
            StrType.BZString,
        ],
    ) -> bytes: ...

    @staticmethod
    def dump(value: list[str] | str, type: StrType) -> bytes:
        """
        Dumps a string of the specified type to a byte array.

        Args:
            value (list[str] | str): String or list of strings to dump.
            type (StrType): String type.

        Returns:
            bytes: Byte array.
        """

        match type:
            case String.StrType.Char | String.StrType.WChar | String.StrType.String:
                if not isinstance(value, str):
                    raise TypeError("'value' must be a string!")

                return value.encode(String.ENCODING)

            case String.StrType.WString:
                if not isinstance(value, str):
                    raise TypeError("'value' must be a string!")

                text = value.encode(String.ENCODING)
                size = Integer.dump(len(text), Integer.IntType.UInt16)
                return size + text

            case String.StrType.BString:
                if not isinstance(value, str):
                    raise TypeError("'value' must be a string!")

                text = value.encode(String.ENCODING)
                size = Integer.dump(len(text), Integer.IntType.UInt8)
                return size + text

            case String.StrType.BZString:
                if not isinstance(value, str):
                    raise TypeError("'value' must be a string!")

                text = value.encode(String.ENCODING) + b"\x00"
                size = Integer.dump(len(text), Integer.IntType.UInt8)
                return size + text

            case String.StrType.List:
                if not isinstance(value, list):
                    raise TypeError("'value' must be a list!")

                data = b"\x00".join(v.encode(String.ENCODING) for v in value) + b"\x00"

                return data


class Flags(IntFlag):
    """
    Class for all types of flags.
    """

    @classmethod
    def parse(cls, data: Stream | bytes, type: Integer.IntType) -> Self:
        """
        Parses a flag from a stream or byte array.

        Args:
            data (Stream | bytes): Stream or byte array.
            type (Integer.IntType): Integer type.

        Returns:
            Self: Parsed flag.
        """

        value: int = Integer.parse(data, type)
        flag = cls(value)

        return flag

    def dump(self, type: Integer.IntType) -> bytes:
        """
        Dumps a flag to a byte array.

        Args:
            type (Integer.IntType): Integer type.

        Returns:
            bytes: Byte array.
        """

        return Integer.dump(self.value, type)


class Hex:
    """
    Class for all types of hexadecimal strings.
    """

    @staticmethod
    def parse(data: Stream | bytes, size: int) -> str:
        return read_data(data, size).hex()

    @staticmethod
    def dump(value: str, type: Integer.IntType) -> bytes:
        number = int(value, base=16)

        return Integer.dump(number, type)


class Hash:
    """
    Class for all types of hashes.
    """

    @staticmethod
    def parse(data: Stream | bytes) -> int:
        return Integer.parse(data, Integer.IntType.UInt64)

    @staticmethod
    def dump(value: int) -> bytes:
        return Integer.dump(value, Integer.IntType.UInt64)

    @staticmethod
    def calc_hash(filename: str) -> int:
        """
        Returns TES4's two hash values for filename.
        Based on TimeSlips code, fixed for names < 3 characters
        and updated to Python 3.

        This original code is from here:
        https://en.uesp.net/wiki/Oblivion_Mod:Hash_Calculation
        """

        name, ext = os.path.splitext(
            filename.lower()
        )  # --"bob.dds" >> root = "bob", ext = ".dds"

        # Create the hashBytes array equivalent
        hash_bytes = [
            ord(name[-1]) if len(name) > 0 else 0,
            ord(name[-2]) if len(name) >= 3 else 0,
            len(name),
            ord(name[0]) if len(name) > 0 else 0,
        ]

        # Convert the byte array to a single 32-bit integer
        hash1: int = struct.unpack("I", bytes(hash_bytes))[0]

        # Apply extensions-specific bit manipulation
        if ext == ".kf":
            hash1 |= 0x80
        elif ext == ".nif":
            hash1 |= 0x8000
        elif ext == ".dds":
            hash1 |= 0x8080
        elif ext == ".wav":
            hash1 |= 0x80000000

        hash2 = 0
        for i in range(1, len(name) - 2):
            hash2 = hash2 * 0x1003F + ord(name[i])

        hash3 = 0
        for char in ext:
            hash3 = hash3 * 0x1003F + ord(char)

        uint_mask = 0xFFFFFFFF
        combined_hash = ((hash2 + hash3) & uint_mask) << 32 | hash1

        return combined_hash



"""
Copyright (c) Cutleast
"""

import os
from io import BufferedReader, BytesIO
from pathlib import Path
from typing import TypeAlias

Stream: TypeAlias = BufferedReader | BytesIO
"""
Type alias for `BufferedReader` and `BytesIO`.
"""


def create_folder_list(folder: Path) -> list[Path]:
    """
    Creates a list with all files
    with relative paths to `folder` and returns it.
    """

    files: list[Path] = []

    for root, _, _files in os.walk(folder):
        for f in _files:
            path = os.path.join(root, f)
            files.append(Path(path).relative_to(folder.parent))

    return files


def get_stream(data: BytesIO | BufferedReader | bytes) -> Stream:
    if isinstance(data, bytes):
        return BytesIO(data)

    return data


def read_data(data: Stream | bytes, size: int) -> bytes:
    """
    Returns `size` bytes from `data`.
    """

    if isinstance(data, bytes):
        return data[:size]
    else:
        return data.read(size)



"""
Copyright (c) Cutleast
"""

from pathlib import Path

from src.sse_bsa import BSAArchive, Header


class TestBSAArchive:
    """
    Tests `bsa_archive.BSAArchive`.

    TODO: Add more tests
    """

    def test_create_archive(self) -> None:
        """
        Tests `bsa_archive.BSAArchive.create_archive` with a txt and a png file.
        """

        # given
        input_folder_path: Path = Path("tests/data")
        txt_file_path: Path = (
            input_folder_path / "interface" / "translations" / "test.txt"
        )
        png_file_path: Path = input_folder_path / "textures" / "test" / "test.png"
        bsa_file_path: Path = Path("tests/data/test.bsa")
        archive_flags: Header.ArchiveFlags = Header.ArchiveFlags(
            Header.ArchiveFlags.EmbedFileNames
        )

        # when
        assert txt_file_path.is_file()
        assert png_file_path.is_file()
        BSAArchive.create_archive(input_folder_path, bsa_file_path, archive_flags)
        bsa_archive: BSAArchive = BSAArchive(bsa_file_path)

        # then
        assert "interface/translations/test.txt" in bsa_archive.files
        assert "textures/test/test.png" in bsa_archive.files
        assert (
            bsa_archive.get_file_stream("interface/translations/test.txt").read()
            == txt_file_path.read_bytes()
        )
        assert (
            bsa_archive.get_file_stream("textures/test/test.png").read()
            == png_file_path.read_bytes()
        )
        assert Header.ArchiveFlags.EmbedFileNames in bsa_archive.header.archive_flags
        assert Header.FileFlags.Menus in bsa_archive.header.file_flags
        assert Header.FileFlags.Textures in bsa_archive.header.file_flags

        bsa_file_path.unlink(missing_ok=True)
